@@ -1,12 +1,13 @@
-from cleo import option
+from __future__ import annotations
 
-from poetry.utils._compat import Path
+from pathlib import Path
 
-from .command import Command
+from cleo.helpers import option
+
+from poetry.console.commands.command import Command
 
 
 class PublishCommand(Command):
-
     name = "publish"
     description = "Publishes a package to a remote repository."
 
@@ -27,6 +28,11 @@ class PublishCommand(Command):
         ),
         option("build", None, "Build the package before publishing."),
         option("dry-run", None, "Perform all actions except upload the package."),
+        option(
+            "skip-existing",
+            None,
+            "Ignore errors from files already existing in the repository.",
+        ),
     ]
 
     help = """The publish command builds and uploads the package to a remote repository.
@@ -38,23 +44,22 @@ The --repository option should match the name of a configured repository using
 the config command.
 """
 
-    loggers = ["poetry.masonry.publishing.publisher"]
+    loggers = ["poetry.publishing.publisher"]
 
-    def handle(self):
+    def handle(self) -> int:
         from poetry.publishing.publisher import Publisher
 
         publisher = Publisher(self.poetry, self.io)
 
         # Building package first, if told
         if self.option("build"):
-            if publisher.files:
-                if not self.confirm(
-                    "There are <info>{}</info> files ready for publishing. "
-                    "Build anyway?".format(len(publisher.files))
-                ):
-                    self.line_error("<error>Aborted!</error>")
+            if publisher.files and not self.confirm(
+                f"There are <info>{len(publisher.files)}</info> files ready for"
+                " publishing. Build anyway?"
+            ):
+                self.line_error("<error>Aborted!</error>")
 
-                    return 1
+                return 1
 
             self.call("build")
 
@@ -81,4 +86,7 @@ the config command.
             cert,
             client_cert,
             self.option("dry-run"),
+            self.option("skip-existing"),
         )
+
+        return 0
